/*
* Copyright (c) 2008-2017, NVIDIA CORPORATION. All rights reserved.
*
* NVIDIA CORPORATION and its licensors retain all intellectual property
* and proprietary rights in and to this software, related documentation
* and any modifications thereto. Any use, reproduction, disclosure or
* distribution of this software and related documentation without an express
* license agreement from NVIDIA CORPORATION is strictly prohibited.
*/


#include "ocean_renderer.h"
#include "test_util.h"

#if USE_VK
	#include <shaders/bytecode/OceanSurfaceVS_vk.h>
	#include <shaders/bytecode/OceanSurfaceHS_vk.h>
	#include <shaders/bytecode/OceanSurfaceDS_vk.h>
	#include <shaders/bytecode/OceanSurfacePS_vk.h>
	#include <shaders/bytecode/MarkerVS_vk.h>
	#include <shaders/bytecode/MarkerPS_vk.h>
#else
	#include <shaders/bytecode/OceanSurfaceVS_dx.h>
	#include <shaders/bytecode/OceanSurfaceHS_dx.h>
	#include <shaders/bytecode/OceanSurfaceDS_dx.h>
	#include <shaders/bytecode/OceanSurfacePS_dx.h>
	#include <shaders/bytecode/MarkerVS_dx.h>
	#include <shaders/bytecode/MarkerPS_dx.h>
#endif // USE_VK

#if USE_DX11
	#define NATIVE_OBJECT (nvrhi::ObjectTypes::D3D11_Resource)
#elif USE_DX12
	#define NATIVE_OBJECT (nvrhi::ObjectTypes::D3D12_Resource)
#elif USE_VK
	#define NATIVE_OBJECT (nvrhi::ObjectTypes::VK_Image)
#endif




struct OCEAN_VS_HS_DS_CBUFFER
{
	// Data used in vertex shader
	gfsdk_float4x4 g_matViewProj;
	gfsdk_float3   g_eyePos;
	float          g_meanOceanLevel;
	// x16 bytes boundary
	float          g_useDiamondPattern;

	// Data used in hull shader
	float          g_dynamicTesselationAmount;
	float          g_staticTesselationOffset;

	// Data used in domain shader
	// Wind waves related data
	float          g_cascade0UVScale;
	// x16 bytes boundary
	float          g_cascade1UVScale;
	float          g_cascade2UVScale;
	float          g_cascade3UVScale;
	float          g_cascade0UVOffset;
	// x16 bytes boundary
	float          g_cascade1UVOffset;
	float          g_cascade2UVOffset;
	float          g_cascade3UVOffset;
	float          g_UVWarpingAmplitude;
	// x16 bytes boundary
	float          g_UVWarpingFrequency;

	// Local waves related data
	float          g_localWavesSimulationDomainWorldspaceSize;
	gfsdk_float2   g_localWavesSimulationDomainWorldspaceCenter;
};

struct OCEAN_PS_CBUFFER
{
	// Defined by wind waves simulation
	float g_cascadeToCascadeScale;
	float g_windWavesTextureSizeInTexels;
	float g_UVWarpingAmplitude;
	float g_UVWarpingFrequency;
											// x16 bytes boundary
	float g_windWavesFoamWhitecapsThreshold;

	// Defined by local waves simulation
	float g_localWavesTextureSizeInTexels;
	float g_localWavesFoamWhitecapsThreshold;
	float g_SimulationDomainSize;
											// x16 bytes boundary
	gfsdk_float2 g_SimulationDomainCenter;

	// Defined by application
	float g_beckmannRoughness;
	float g_showCascades;
											// x16 bytes boundary
	gfsdk_float3 g_sunDirection;
	float g_sunIntensity;
											// x16 bytes boundary
	gfsdk_float3 g_eyePos;
	float g_useMicrofacetFresnel;
											// x16 bytes boundary
	float g_useMicrofacetSpecular;
	float g_useMicrofacetReflection;

	// CBs must be multiple of 16 bytes large
	float pad0;
	float pad1;
};

struct OCEAN_VS_CBUFFER_PERINSTANCE_ENTRY
{
	gfsdk_float2 g_patchWorldspaceOrigin;
	float g_patchWorldspaceScale;
	float g_patchMorphConstantAndSign;
};

struct MARKER_VS_CBUFFER
{
	gfsdk_float4x4 g_matViewProj;
};

uint32_t getNumMipLevels(uint32_t const squareTextureSize)
{
	uint32_t mipLevels = 1;
	uint32_t texSize = squareTextureSize;
	while (texSize > 1)
	{
		mipLevels++;
		texSize /= 2;
	}
	return mipLevels;
}

bool OceanRenderer::Init()
{
	nvrhi::CommandListParameters cmdListParams;
	m_hCommandList = m_pDevice->createCommandList(cmdListParams);
	m_hCommandList->open();

	// Creating depth buffer 
	// it is created already if BackBufferResized was called, so destroying it first then
	if (m_pState->hDepthBufferTexture != nullptr)
	{
		m_pState->hDepthBufferTexture = nullptr;
	}
	nvrhi::TextureDesc td;
	td.width = m_backBufferWidth;
	td.height = m_backBufferHeight;
	td.sampleCount = m_sampleCount;
	td.debugName = "Depth buffer created in OceanRenderer::Init()";
	td.format = nvrhi::Format::D24S8;
	td.isRenderTarget = true;
	td.clearValue = nvrhi::Color(1.0f);
	td.useClearValue = true;
	td.isTypeless = true; // Depth resources must be declared typeless in NVRHI
	m_pState->hDepthBufferTexture = m_pDevice->createTexture(td);
	m_hCommandList->beginTrackingTextureState(m_pState->hDepthBufferTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::COMMON);
	m_hCommandList->endTrackingTextureState(m_pState->hDepthBufferTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::DEPTH_WRITE);

	// Initializing frame buffer #0
	td.debugName = "proxy color target";
#if USE_VK
	td.format = nvrhi::Format::BGRA8_UNORM;
#else
	td.format = nvrhi::Format::RGBA8_UNORM;
#endif
	m_hInitTimeRenderTargetTexture = m_pDevice->createTexture(td);
	nvrhi::FramebufferDesc fbDesc;
	fbDesc.addColorAttachment(m_hInitTimeRenderTargetTexture, 0, 0);
	fbDesc.setDepthAttachment(m_pState->hDepthBufferTexture, 0, 0);
	m_hInitTimeFB = m_pDevice->createFramebuffer(fbDesc);


	// Loading shaders
#if USE_VK
	auto VSOceanBinary = OceanSurfaceVS_bytecode_vk;
	auto VSOceanBinarySize = sizeof(OceanSurfaceVS_bytecode_vk);
	auto HSOceanBinary = OceanSurfaceHS_bytecode_vk;
	auto HSOceanBinarySize = sizeof(OceanSurfaceHS_bytecode_vk);
	auto DSOceanBinary = OceanSurfaceDS_bytecode_vk;
	auto DSOceanBinarySize = sizeof(OceanSurfaceDS_bytecode_vk);
	auto PSOceanBinary = OceanSurfacePS_bytecode_vk;
	auto PSOceanBinarySize = sizeof(OceanSurfacePS_bytecode_vk);

	auto VSOceanMarkerBinary = MarkerVS_bytecode_vk;
	auto VSOceanMarkerBinarySize = sizeof(MarkerVS_bytecode_vk);
	auto PSOceanMarkerBinary = MarkerPS_bytecode_vk;
	auto PSOceanMarkerBinarySize = sizeof(MarkerPS_bytecode_vk);
#else 
	auto VSOceanBinary = OceanSurfaceVS_bytecode_dx;
	auto VSOceanBinarySize = sizeof(OceanSurfaceVS_bytecode_dx);
	auto HSOceanBinary = OceanSurfaceHS_bytecode_dx;
	auto HSOceanBinarySize = sizeof(OceanSurfaceHS_bytecode_dx);
	auto DSOceanBinary = OceanSurfaceDS_bytecode_dx;
	auto DSOceanBinarySize = sizeof(OceanSurfaceDS_bytecode_dx);
	auto PSOceanBinary = OceanSurfacePS_bytecode_dx;
	auto PSOceanBinarySize = sizeof(OceanSurfacePS_bytecode_dx);

	auto VSOceanMarkerBinary = MarkerVS_bytecode_dx;
	auto VSOceanMarkerBinarySize = sizeof(MarkerVS_bytecode_dx);
	auto PSOceanMarkerBinary = MarkerPS_bytecode_dx;
	auto PSOceanMarkerBinarySize = sizeof(MarkerPS_bytecode_dx);
#endif 


	// Creating shaders
	m_hOceanSurfaceVS = m_pDevice->createShader(nvrhi::ShaderDesc(nvrhi::ShaderType::SHADER_VERTEX), VSOceanBinary, VSOceanBinarySize);
	if (m_hOceanSurfaceVS == nullptr) return false;

	m_hOceanSurfaceHS = m_pDevice->createShader(nvrhi::ShaderDesc(nvrhi::ShaderType::SHADER_HULL), HSOceanBinary, HSOceanBinarySize);
	if (m_hOceanSurfaceHS == nullptr) return false;

	m_hOceanSurfaceDS = m_pDevice->createShader(nvrhi::ShaderDesc(nvrhi::ShaderType::SHADER_DOMAIN), DSOceanBinary, DSOceanBinarySize);
	if (m_hOceanSurfaceDS == nullptr) return false;

	m_hOceanSurfacePS = m_pDevice->createShader(nvrhi::ShaderDesc(nvrhi::ShaderType::SHADER_PIXEL), PSOceanBinary, PSOceanBinarySize);
	if (m_hOceanSurfacePS == nullptr) return false;

	m_hMarkerVS = m_pDevice->createShader(nvrhi::ShaderDesc(nvrhi::ShaderType::SHADER_VERTEX), VSOceanMarkerBinary, VSOceanMarkerBinarySize);
	if (m_hMarkerVS == nullptr) return false;

	m_hMarkerPS = m_pDevice->createShader(nvrhi::ShaderDesc(nvrhi::ShaderType::SHADER_PIXEL), PSOceanMarkerBinary, PSOceanMarkerBinarySize);
	if (m_hMarkerPS == nullptr) return false;


	// Creating input layout for ocean surface rendering 
	nvrhi::VertexAttributeDesc oceanSurceaceLayoutDesc[] = {
		{ "POSITION", nvrhi::Format::RG32_FLOAT, 1, 0, 0, 8, false },
	};

	m_hOceanSurfaceLayout = m_pDevice->createInputLayout(oceanSurceaceLayoutDesc, 1, m_hOceanSurfaceVS);

	// Creating ocean surface vertex and index buffers
	uint32_t numVertices = 0;
	uint32_t numIndices = 0;
	gfsdk_float2 *pVertexData = nullptr;
	uint32_t *pIndexData = nullptr;
	std::vector<gfsdk_float2> oceanVertices;
	std::vector<uint32_t> oceanIndices;
	
	GFSDK_WaveWorks_Quadtree_GetGeometryData(m_pState->hOceanQuadtree, &numVertices, &pVertexData, &numIndices, &pIndexData);

	oceanVertices.assign(pVertexData, pVertexData + numVertices);
	oceanIndices.assign(pIndexData, pIndexData + numIndices);

	nvrhi::BufferDesc bd;
	bd.byteSize = numVertices * sizeof(gfsdk_float2);
	bd.isVertexBuffer = true;
	bd.isIndexBuffer = false;
	m_hOceanSurfaceVB = m_pDevice->createBuffer(bd);
	m_hCommandList->beginTrackingBufferState(m_hOceanSurfaceVB, nvrhi::ResourceStates::COMMON);
	m_hCommandList->writeBuffer(m_hOceanSurfaceVB, oceanVertices.data(), bd.byteSize);
	m_hCommandList->endTrackingBufferState(m_hOceanSurfaceVB, nvrhi::ResourceStates::VERTEX_BUFFER, true);

	bd.byteSize = numIndices * sizeof(uint32_t);
	bd.isVertexBuffer = false;
	bd.isIndexBuffer = true;
	m_hOceanSurfaceIB = m_pDevice->createBuffer(bd);
	m_hCommandList->beginTrackingBufferState(m_hOceanSurfaceIB, nvrhi::ResourceStates::COMMON);
	m_hCommandList->writeBuffer(m_hOceanSurfaceIB, oceanIndices.data(), bd.byteSize);
	m_hCommandList->endTrackingBufferState(m_hOceanSurfaceIB, nvrhi::ResourceStates::INDEX_BUFFER, true);

	// Creating textures
	m_hFoamIntensityTexture = CreateTextureFromFile(m_pDevice, m_hCommandList, (GetMediaFolder() / "foam_intensity.dds").generic_string().c_str());
	m_hFoamBubblesTexture = CreateTextureFromFile(m_pDevice, m_hCommandList, (GetMediaFolder() / "foam_bubbles.dds").generic_string().c_str());
	m_hWindGustsTexture = CreateTextureFromFile(m_pDevice, m_hCommandList, (GetMediaFolder() / "wind_gusts.dds").generic_string().c_str());

	if (m_hFoamIntensityTexture == nullptr) return false;
	if (m_hFoamBubblesTexture == nullptr) return false;
	if (m_hWindGustsTexture == nullptr) return false;

	// Creating samplers
	nvrhi::SamplerDesc samplerDesc;
	samplerDesc.magFilter = true;
	samplerDesc.minFilter = true;
	samplerDesc.mipFilter = true;
	samplerDesc.wrapMode[0] = nvrhi::SamplerDesc::WrapMode::WRAP_MODE_WRAP;
	samplerDesc.wrapMode[1] = nvrhi::SamplerDesc::WrapMode::WRAP_MODE_WRAP;
	samplerDesc.wrapMode[2] = nvrhi::SamplerDesc::WrapMode::WRAP_MODE_WRAP;
	m_hBilinearSampler = m_pDevice->createSampler(samplerDesc);

	samplerDesc.anisotropy = 4.0f;
	m_hAnisotropicSampler = m_pDevice->createSampler(samplerDesc);

	samplerDesc.anisotropy = 1.0f;
	samplerDesc.wrapMode[0] = nvrhi::SamplerDesc::WrapMode::WRAP_MODE_CLAMP;
	samplerDesc.wrapMode[1] = nvrhi::SamplerDesc::WrapMode::WRAP_MODE_CLAMP;
	samplerDesc.wrapMode[2] = nvrhi::SamplerDesc::WrapMode::WRAP_MODE_CLAMP;
	m_hBilinearClampToEdgeSampler = m_pDevice->createSampler(samplerDesc);

	// Creating NVRHI textures from native objects exposed from WaveWorks
	GFSDK_WaveWorks_Wind_Waves_Rendering_Data windWavesRenderingData;
	GFSDK_WaveWorks_Local_Waves_Rendering_Data localWavesRenderingData;
	GFSDK_WaveWorks_Wind_Waves_GetDataForRendering(m_pState->hOceanWindSimulation, windWavesRenderingData);
	GFSDK_WaveWorks_Local_Waves_GetDataForRendering(m_pState->hOceanLocalSimulation, localWavesRenderingData);

	nvrhi::TextureDesc texArrayDesc;
	texArrayDesc.width = windWavesRenderingData.size_of_texture_arrays;
	texArrayDesc.height = windWavesRenderingData.size_of_texture_arrays;
	texArrayDesc.format = nvrhi::Format::RGBA16_FLOAT;
	texArrayDesc.isUAV = true;
	texArrayDesc.dimension = nvrhi::TextureDimension::Texture2DArray;
	texArrayDesc.arraySize = 4;
	texArrayDesc.mipLevels = getNumMipLevels(windWavesRenderingData.size_of_texture_arrays);
	texArrayDesc.initialState = nvrhi::ResourceStates::SHADER_RESOURCE;
	texArrayDesc.keepInitialState = true;
	m_hWindWavesDisplacementsTextureArray = m_pDevice->createHandleForNativeTexture(NATIVE_OBJECT, windWavesRenderingData.displacements_texture_array, texArrayDesc);
	m_hWindWavesGradientsTextureArray = m_pDevice->createHandleForNativeTexture(NATIVE_OBJECT, windWavesRenderingData.gradients_texture_array, texArrayDesc);
	m_hWindWavesMomentsTextureArray = m_pDevice->createHandleForNativeTexture(NATIVE_OBJECT, windWavesRenderingData.moments_texture_array, texArrayDesc);

	nvrhi::TextureDesc texDesc;
	texDesc.width = localWavesRenderingData.size_of_texture;
	texDesc.height = localWavesRenderingData.size_of_texture;
	texDesc.format = nvrhi::Format::RGBA16_FLOAT;
	texDesc.isUAV = true;
	texDesc.dimension = nvrhi::TextureDimension::Texture2D;
	texDesc.mipLevels = getNumMipLevels(localWavesRenderingData.size_of_texture);
	texDesc.initialState = nvrhi::ResourceStates::SHADER_RESOURCE;
	texDesc.keepInitialState = true;
	m_hLocalWavesDisplacementsTexture = m_pDevice->createHandleForNativeTexture(NATIVE_OBJECT, localWavesRenderingData.displacements_texture, texDesc);
	m_hLocalWavesGradientsTexture = m_pDevice->createHandleForNativeTexture(NATIVE_OBJECT, localWavesRenderingData.gradients_texture, texDesc);
		
	// Creating constant buffer for vertex/domain shader
	nvrhi::BufferDesc cbDesc;
	cbDesc.byteSize = sizeof(OCEAN_VS_HS_DS_CBUFFER);
	cbDesc.debugName = "Ocean VS/HS/DS CB";
	cbDesc.isConstantBuffer = true;
	cbDesc.isVolatile = false;
	m_hOceanSurfaceVSHSDSCB = m_pDevice->createBuffer(cbDesc);

	// Creating constant buffer for pixel shader
	cbDesc.byteSize = sizeof(OCEAN_PS_CBUFFER);
	cbDesc.debugName = "Ocean PS CB";
	cbDesc.isConstantBuffer = true;
	cbDesc.isVolatile = false;
	m_hOceanSurfacePSCB = m_pDevice->createBuffer(cbDesc);

	// Creating perinstance constant buffer for vertex shader
	cbDesc.byteSize = 4096*sizeof(OCEAN_VS_CBUFFER_PERINSTANCE_ENTRY); // 64KB or 4096*4 floats is maximum amout of data allowed for constant buffers
	cbDesc.debugName = "Ocean VS per instance CB";
	cbDesc.isConstantBuffer = true;
	cbDesc.isVolatile = true;
	m_hOceanSurfacePerInstanceCB = m_pDevice->createBuffer(cbDesc);

	// Creating ocean draw call state members
	nvrhi::BlendState blendState;
	blendState.blendEnable[0] = false;

	nvrhi::RasterState rasterState;
	rasterState.fillMode = nvrhi::RasterState::FILL_SOLID;
	rasterState.cullMode = nvrhi::RasterState::CULL_FRONT;
	rasterState.scissorEnable = false;
	rasterState.depthClipEnable = true;

	nvrhi::DepthStencilState depthStencilState;
	depthStencilState.depthEnable = true;
	depthStencilState.depthWriteMask = nvrhi::DepthStencilState::DEPTH_WRITE_MASK_ALL;
	depthStencilState.depthFunc = nvrhi::DepthStencilState::COMPARISON_LESS_EQUAL;
	depthStencilState.stencilEnable = false;


	nvrhi::RenderState renderState;
	renderState.blendState = blendState;
	renderState.depthStencilState = depthStencilState;
	renderState.rasterState = rasterState;

	nvrhi::GraphicsPipelineDesc psoDesc;
	nvrhi::BindingLayoutDesc blDesc;
	nvrhi::BindingSetDesc bsDesc;

	nvrhi::TextureDesc tdWindDisplacements = m_hWindWavesDisplacementsTextureArray->GetDesc();
	nvrhi::TextureDesc tdLocalDisplacements = m_hLocalWavesDisplacementsTexture->GetDesc();
	nvrhi::TextureDesc tdFoamBubbles = m_hFoamBubblesTexture->GetDesc();
	nvrhi::TextureDesc tdFoamIntensity = m_hFoamIntensityTexture->GetDesc();
	nvrhi::TextureDesc tdWindGusts = m_hWindGustsTexture->GetDesc();

	blDesc.VS = { { 0, nvrhi::ResourceType::VolatileConstantBuffer, 0 },
					{ 1, nvrhi::ResourceType::ConstantBuffer, 0 } };
	blDesc.HS = { { 0, nvrhi::ResourceType::ConstantBuffer, 0 } };
	blDesc.DS = { { 0, nvrhi::ResourceType::Texture_SRV, 0 },
					{ 1, nvrhi::ResourceType::Texture_SRV, 0 },
					{ 0, nvrhi::ResourceType::Sampler, 0 },
					{ 1, nvrhi::ResourceType::Sampler, 0 },
					{ 0, nvrhi::ResourceType::ConstantBuffer, 0 } };
	blDesc.PS = { { 0, nvrhi::ResourceType::Texture_SRV, 0 },
				   { 1, nvrhi::ResourceType::Texture_SRV, 0 },
				   { 2, nvrhi::ResourceType::Texture_SRV, 0 },
				   { 3, nvrhi::ResourceType::Texture_SRV, 0 },
				   { 4, nvrhi::ResourceType::Texture_SRV, 0 },
				   { 5, nvrhi::ResourceType::Texture_SRV, 0 },
				   { 6, nvrhi::ResourceType::Texture_SRV, 0 },
				   { 7, nvrhi::ResourceType::Texture_SRV, 0 },
				   { 0, nvrhi::ResourceType::Sampler, 0 },
				   { 1, nvrhi::ResourceType::Sampler, 0 },
				   { 2, nvrhi::ResourceType::Sampler, 0 },
				   { 0, nvrhi::ResourceType::ConstantBuffer, 0 } };
	m_hOceanSurfaceBindingLayout = m_pDevice->createBindingLayout(blDesc);

	bsDesc.VS = { nvrhi::BindingSetItem::ConstantBuffer(0, m_hOceanSurfacePerInstanceCB ),
					nvrhi::BindingSetItem::ConstantBuffer(1, m_hOceanSurfaceVSHSDSCB ) };
	bsDesc.HS = { nvrhi::BindingSetItem::ConstantBuffer(0, m_hOceanSurfaceVSHSDSCB ) };
	bsDesc.DS = { nvrhi::BindingSetItem::Texture_SRV(0, m_hWindWavesDisplacementsTextureArray, nvrhi::Format::RGBA16_FLOAT, nvrhi::TextureSubresourceSet(0,tdWindDisplacements.mipLevels,0,tdWindDisplacements.arraySize) ),
					nvrhi::BindingSetItem::Texture_SRV(1, m_hLocalWavesDisplacementsTexture, nvrhi::Format::RGBA16_FLOAT),
					nvrhi::BindingSetItem::Sampler(0, m_hBilinearSampler ),
					nvrhi::BindingSetItem::Sampler(1, m_hBilinearClampToEdgeSampler ),
					nvrhi::BindingSetItem::ConstantBuffer(0, m_hOceanSurfaceVSHSDSCB) };
	bsDesc.PS = { nvrhi::BindingSetItem::Texture_SRV(0, m_hWindWavesGradientsTextureArray, nvrhi::Format::RGBA16_FLOAT, nvrhi::TextureSubresourceSet(0,tdWindDisplacements.mipLevels,0,tdWindDisplacements.arraySize)),
					nvrhi::BindingSetItem::Texture_SRV(1, m_hWindWavesMomentsTextureArray, nvrhi::Format::RGBA16_FLOAT, nvrhi::TextureSubresourceSet(0,tdWindDisplacements.mipLevels,0,tdWindDisplacements.arraySize)),
					nvrhi::BindingSetItem::Texture_SRV(2, m_hLocalWavesDisplacementsTexture,  nvrhi::Format::RGBA16_FLOAT),
					nvrhi::BindingSetItem::Texture_SRV(3, m_hLocalWavesGradientsTexture,  nvrhi::Format::RGBA16_FLOAT),
					nvrhi::BindingSetItem::Texture_SRV(4, m_hFoamIntensityTexture, nvrhi::Format::RGBA8_UNORM),
					nvrhi::BindingSetItem::Texture_SRV(5, m_hFoamBubblesTexture,  nvrhi::Format::RGBA8_UNORM),
					nvrhi::BindingSetItem::Texture_SRV(6, m_hWindGustsTexture,  nvrhi::Format::RGBA8_UNORM),
					nvrhi::BindingSetItem::Texture_SRV(7, m_pState->hDynamicSkyDomeTexture,  nvrhi::Format::RGBA16_FLOAT),
					nvrhi::BindingSetItem::Sampler(0, m_hBilinearSampler),
					nvrhi::BindingSetItem::Sampler(1, m_hAnisotropicSampler),
					nvrhi::BindingSetItem::Sampler(2, m_hBilinearClampToEdgeSampler ),
					nvrhi::BindingSetItem::ConstantBuffer(0, m_hOceanSurfacePSCB ) };
	m_hOceanSurfaceBindings = m_pDevice->createBindingSet(bsDesc, m_hOceanSurfaceBindingLayout);

	psoDesc.VS = m_hOceanSurfaceVS;
	psoDesc.HS = m_hOceanSurfaceHS;
	psoDesc.DS = m_hOceanSurfaceDS;
	psoDesc.PS = m_hOceanSurfacePS;
	psoDesc.primType = nvrhi::PrimitiveType::PATCH_3_CONTROL_POINT;
	psoDesc.bindingLayouts = { m_hOceanSurfaceBindingLayout };
	psoDesc.inputLayout = m_hOceanSurfaceLayout;
	psoDesc.renderState = renderState;
	psoDesc.renderState.rasterState.fillMode = nvrhi::RasterState::FILL_SOLID;
	m_hOceanSurfaceSolidPSO = m_pDevice->createGraphicsPipeline(psoDesc, m_hInitTimeFB);

	psoDesc.renderState.rasterState.fillMode = nvrhi::RasterState::FILL_LINE;
	m_hOceanSurfaceWireframePSO = m_pDevice->createGraphicsPipeline(psoDesc, m_hInitTimeFB);

	// Allocating memory for readback positions and results
	m_pReadbackPositions = (gfsdk_float2*)malloc(NumMarkers * sizeof(gfsdk_float2));
	m_pReadbackResults = (gfsdk_float4*)malloc(NumMarkers * sizeof(gfsdk_float4));

	// Creating constant buffer for marker rendering
	cbDesc.byteSize = sizeof(MARKER_VS_CBUFFER);
	cbDesc.debugName = "VS CB for rendering markers";
	cbDesc.isConstantBuffer = true;
	m_hMarkerVSCB = m_pDevice->createBuffer(cbDesc);
	m_hCommandList->beginTrackingBufferState(m_hMarkerVSCB, nvrhi::ResourceStates::COMMON);
	m_hCommandList->beginTrackingBufferState(m_hMarkerVSCB, nvrhi::ResourceStates::CONSTANT_BUFFER);

	// Creating input layout for rendering markers
	nvrhi::VertexAttributeDesc markers_layout[] = {
		{ "POSITION", nvrhi::Format::RGBA32_FLOAT, 1, 0, 0, 16, false },
		{ "PERINSTANCEDATA", nvrhi::Format::RGBA32_FLOAT, 1, 1, 0, 16, true },
	};

	m_hMarkerLayout = m_pDevice->createInputLayout(markers_layout, 2, m_hMarkerVS);

	// Creating PSO and bindings marker rendering
	blDesc.VS = { { 0, nvrhi::ResourceType::VolatileConstantBuffer, 0 } };
	blDesc.HS = {};
	blDesc.DS = {};
	blDesc.PS = {};
	m_hMarkerBindingLayout = m_pDevice->createBindingLayout(blDesc);

	bsDesc.VS = { nvrhi::BindingSetItem::ConstantBuffer(0, m_hMarkerVSCB ) };
	bsDesc.HS = {};
	bsDesc.DS = {};
	bsDesc.PS = {};
	m_hMarkerBindings = m_pDevice->createBindingSet(bsDesc, m_hMarkerBindingLayout);

	psoDesc.VS = m_hMarkerVS;
	psoDesc.HS = nullptr;
	psoDesc.DS = nullptr;
	psoDesc.PS = m_hMarkerPS;
	psoDesc.renderState.rasterState.fillMode = nvrhi::RasterState::FILL_SOLID;
	psoDesc.renderState.rasterState.cullMode = nvrhi::RasterState::CULL_NONE;
	psoDesc.primType = nvrhi::PrimitiveType::TRIANGLE_LIST;
	psoDesc.bindingLayouts = { m_hMarkerBindingLayout };
	psoDesc.inputLayout = m_hMarkerLayout;

	m_hMarkerPSO = m_pDevice->createGraphicsPipeline(psoDesc, m_hInitTimeFB);


	// Creating marker instance buffer and upload buffers for marker instances
	nvrhi::BufferDesc desc;
	desc.byteSize = uint32_t(NumMarkers * sizeof(gfsdk_float4));
	desc.structStride = 0;
	desc.debugName = "Marker instances upload buffer";
	desc.canHaveUAVs = false;
	desc.isVertexBuffer = true;
	desc.isIndexBuffer = false;
	desc.isDrawIndirectArgs = false;
	desc.disableGPUsSync = false;
	desc.cpuAccess = nvrhi::CpuAccessMode::Write;

	for (int i = 0; i < 3; i++)
	{
		m_hMarkerInstanceUploadBuffers[i] = m_pDevice->createBuffer(desc);
		if (!m_hMarkerInstanceUploadBuffers[i])
		{
			return false;
		}
	}
	m_markerInstanceUploadBufferNumber = 0;

	desc.cpuAccess = nvrhi::CpuAccessMode::None;
	m_hMarkerInstanceBuffer = m_pDevice->createBuffer(desc);

	// Creating marker VB and IB
	float marker_vertexdata[20] = { 0.0f, 0.0f, 0.0f, 1.0f,
									0.5f,-0.5f, 1.0f, 1.0f,
									0.5f, 0.5f, 1.0f, 1.0f,
								   -0.5f, 0.5f, 1.0f, 1.0f,
								   -0.5f,-0.5f, 1.0f, 1.0f };
	uint32_t marker_indexdata[12] = { 0, 1, 2,
									 0, 2, 3,
									 0, 3, 4,
									 0, 4, 1 };

	bd.byteSize = 20 * sizeof(float);
	bd.isVertexBuffer = true;
	m_hMarkerVB = m_pDevice->createBuffer(bd);
	m_hCommandList->beginTrackingBufferState(m_hMarkerVB, nvrhi::ResourceStates::COMMON);
	m_hCommandList->writeBuffer(m_hMarkerVB, marker_vertexdata, bd.byteSize);
	m_hCommandList->endTrackingBufferState(m_hMarkerVB, nvrhi::ResourceStates::VERTEX_BUFFER, true);

	bd.byteSize = 12 * sizeof(uint32_t);
	bd.isIndexBuffer = true;
	m_hMarkerIB = m_pDevice->createBuffer(bd);
	m_hCommandList->beginTrackingBufferState(m_hMarkerIB, nvrhi::ResourceStates::COMMON);
	m_hCommandList->writeBuffer(m_hMarkerIB, marker_indexdata, bd.byteSize);
	m_hCommandList->endTrackingBufferState(m_hMarkerIB, nvrhi::ResourceStates::INDEX_BUFFER, true);



	m_hCommandList->close();
	m_pDevice->executeCommandList(m_hCommandList);

	return true;
}

void OceanRenderer::BackBufferResizing() 
{
	m_hFBs.clear();
	m_pState->hDepthBufferTexture = nullptr;
}
void OceanRenderer::BackBufferResized(const uint32_t width, const uint32_t height, const uint32_t sampleCount)
{
	m_backBufferWidth = width;
	m_backBufferHeight = height;
	m_sampleCount = sampleCount;
	
	if (m_pDevice == nullptr) return;
	if (m_pState == nullptr) return;

	m_pDevice->waitForIdle();

	updateCameraProjection();
	m_needToRecreatePSOs = true;
}

void OceanRenderer::updateCameraProjection()
{
	float aspectRatio = (float)m_backBufferWidth / (float)m_backBufferHeight;
	m_pState->Camera.SetProjection(3.14f/3.0f, aspectRatio, m_pState->fCameraClipNear, m_pState->fCameraClipFar);
}

void OceanRenderer::Render(nvrhi::ITexture* renderTarget, uint32_t swapChainIndex)
{
	m_hCommandList->open();
	nvrhi::FramebufferHandle FB = getFramebuffer(renderTarget, swapChainIndex);
	m_hCommandList->beginTrackingTextureState(renderTarget, nvrhi::AllSubresources, nvrhi::ResourceStates::RENDER_TARGET);
	m_hCommandList->beginTrackingTextureState(m_pState->hDepthBufferTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::DEPTH_WRITE);

	// Clearing depth/color RTs
	m_hCommandList->clearTextureFloat(renderTarget, nvrhi::TextureSubresourceSet(0, 1, 0, 1), nvrhi::Color(0.8f, 0.9f, 1.0f, 1.0f));
	m_hCommandList->clearTextureFloat(m_pState->hDepthBufferTexture, nvrhi::TextureSubresourceSet(0, 1, 0, 1), nvrhi::Color(1.0f));


	// Rebinding resources if needed
	if (m_pState->bNeedToUpdateQuadtreeProperties)
	{
		GFSDK_WaveWorks_Quadtree_UpdateParams(m_pState->hOceanQuadtree, m_pState->OceanQuadtreeParameters);
		RebindOceanResources();
		m_pState->bNeedToUpdateQuadtreeProperties = false;
	}

	if (m_pState->bNeedToUpdateLocalWavesSimulationProperties)
	{
		GFSDK_WaveWorks_Local_Waves_Simulation_UpdateProperties(m_pState->hOceanLocalSimulation, m_pState->OceanLocalSimulationSettings, m_pState->OceanLocalSimulationParameters);
		RebindOceanResources();
		m_pState->bNeedToUpdateLocalWavesSimulationProperties = false;
	}

	if (m_pState->bNeedToUpdateWindWavesSimulationProperties)
	{
		GFSDK_WaveWorks_Wind_Waves_Simulation_UpdateProperties(m_pState->hOceanWindSimulation, m_pState->OceanWindSimulationSettings, m_pState->OceanWindSimulationParameters);
		RebindOceanResources();
		m_pState->bNeedToUpdateWindWavesSimulationProperties = false;
	}

	if (m_pState->bRenderWater)
	{
		m_hCommandList->beginTrackingTextureState(m_hWindWavesDisplacementsTextureArray, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
		m_hCommandList->beginTrackingTextureState(m_hWindWavesGradientsTextureArray, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
		m_hCommandList->beginTrackingTextureState(m_hWindWavesMomentsTextureArray, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
		m_hCommandList->beginTrackingTextureState(m_hFoamIntensityTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
		m_hCommandList->beginTrackingTextureState(m_hFoamBubblesTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
		m_hCommandList->beginTrackingTextureState(m_hWindGustsTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
		m_hCommandList->beginTrackingTextureState(m_pState->hDynamicSkyDomeTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
		m_hCommandList->beginTrackingTextureState(m_hLocalWavesDisplacementsTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
		m_hCommandList->beginTrackingTextureState(m_hLocalWavesGradientsTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
		m_hCommandList->beginTrackingBufferState(m_hOceanSurfacePSCB, nvrhi::ResourceStates::CONSTANT_BUFFER);
		m_hCommandList->beginTrackingBufferState(m_hOceanSurfaceVSHSDSCB, nvrhi::ResourceStates::CONSTANT_BUFFER);

		// Initializing ocean surface draw state
		nvrhi::Viewport vp;
		vp.maxX = float(m_backBufferWidth);
		vp.maxY = float(m_backBufferHeight);

		nvrhi::Rect scissor;
		scissor.maxX = m_backBufferWidth;
		scissor.maxY = m_backBufferHeight;
	
		if (m_pState->bRenderWireframe)
		{
			m_drawState.pipeline = m_hOceanSurfaceWireframePSO;
		}
		else
		{
			m_drawState.pipeline = m_hOceanSurfaceSolidPSO;
		}
		m_drawState.framebuffer = FB;
		m_drawState.bindings = { m_hOceanSurfaceBindings }; 
		m_drawState.viewport.viewports.resize(1);
		m_drawState.viewport.viewports[0] = vp;
		m_drawState.viewport.scissorRects = { scissor };
		gfsdk_float2 viewportSize;
		viewportSize.x = float(m_backBufferWidth);
		viewportSize.y = float(m_backBufferHeight);

		// Getting view and projection matrices
		gfsdk_float4x4 viewMatrix = m_pState->Camera.GetWorldToView();
		gfsdk_float4x4 projMatrix = m_pState->Camera.GetProjection();
		gfsdk_float3 eyePoint = m_pState->Camera.GetCameraPosition();

		//Internal quadtree math is left handed, z up, so flippings rows 2 and 3 here
		gfsdk_float4x4  vm = { viewMatrix._11, viewMatrix._12, viewMatrix._13,viewMatrix._14,
					   viewMatrix._31, viewMatrix._32, viewMatrix._33,viewMatrix._34,
					   viewMatrix._21, viewMatrix._22, viewMatrix._23,viewMatrix._24,
					   viewMatrix._41, viewMatrix._42, viewMatrix._43,viewMatrix._44 };

		// Getting data for rendering from WaveWorks 
		GFSDK_WaveWorks_Wind_Waves_Rendering_Data windWavesRenderingData;
		GFSDK_WaveWorks_Wind_Waves_GetDataForRendering(m_pState->hOceanWindSimulation, windWavesRenderingData);

		GFSDK_WaveWorks_Local_Waves_Rendering_Data localWavesRenderingData;
		GFSDK_WaveWorks_Local_Waves_GetDataForRendering(m_pState->hOceanLocalSimulation, localWavesRenderingData);

		// Updating constant buffers
		OCEAN_VS_HS_DS_CBUFFER VSHSDSCB;
		VSHSDSCB.g_matViewProj = projMatrix * vm;
		VSHSDSCB.g_eyePos = { eyePoint.x, eyePoint.z, eyePoint.y };
		VSHSDSCB.g_meanOceanLevel = m_pState->OceanQuadtreeParameters.mean_sea_level;
		VSHSDSCB.g_dynamicTesselationAmount = m_pState->fTessellationAmount;
		VSHSDSCB.g_staticTesselationOffset = m_pState->fTessellationOffset;
		VSHSDSCB.g_cascade0UVOffset = windWavesRenderingData.cascade0_UV_offset;
		VSHSDSCB.g_cascade0UVScale = windWavesRenderingData.cascade0_UV_scale;
		VSHSDSCB.g_cascade1UVOffset = windWavesRenderingData.cascade1_UV_offset;
		VSHSDSCB.g_cascade1UVScale = windWavesRenderingData.cascade1_UV_scale;
		VSHSDSCB.g_cascade2UVOffset = windWavesRenderingData.cascade2_UV_offset;
		VSHSDSCB.g_cascade2UVScale = windWavesRenderingData.cascade2_UV_scale;
		VSHSDSCB.g_cascade3UVOffset = windWavesRenderingData.cascade3_UV_offset;
		VSHSDSCB.g_cascade3UVScale = windWavesRenderingData.cascade3_UV_scale;
		VSHSDSCB.g_localWavesSimulationDomainWorldspaceCenter = localWavesRenderingData.simulation_domain_worldspace_center;
		VSHSDSCB.g_localWavesSimulationDomainWorldspaceSize = localWavesRenderingData.simulation_domain_worldspace_size;
		VSHSDSCB.g_useDiamondPattern = m_pState->OceanQuadtreeParameters.generate_diamond_pattern ? 1.0f : 0.0f;
		VSHSDSCB.g_UVWarpingAmplitude = windWavesRenderingData.uv_warping_amplitude;
		VSHSDSCB.g_UVWarpingFrequency = windWavesRenderingData.uv_warping_frequency;

		m_hCommandList->writeBuffer(m_hOceanSurfaceVSHSDSCB, &VSHSDSCB, sizeof(VSHSDSCB));
		
		OCEAN_PS_CBUFFER PSCB;
		PSCB.g_cascadeToCascadeScale = windWavesRenderingData.cascade_to_cascade_scale;
		PSCB.g_windWavesTextureSizeInTexels = (float)windWavesRenderingData.size_of_texture_arrays;
		PSCB.g_UVWarpingAmplitude = windWavesRenderingData.uv_warping_amplitude;
		PSCB.g_UVWarpingFrequency = windWavesRenderingData.uv_warping_frequency;
		PSCB.g_windWavesFoamWhitecapsThreshold = windWavesRenderingData.foam_whitecaps_threshold;
		PSCB.g_SimulationDomainCenter = localWavesRenderingData.simulation_domain_worldspace_center;
		PSCB.g_SimulationDomainSize = localWavesRenderingData.simulation_domain_worldspace_size;
		PSCB.g_localWavesTextureSizeInTexels = (float)localWavesRenderingData.size_of_texture;
		PSCB.g_localWavesFoamWhitecapsThreshold = m_pState->OceanLocalSimulationParameters.foam_whitecaps_threshold;
		PSCB.g_beckmannRoughness = m_pState->fBeckmannRoughness;
		PSCB.g_sunIntensity = m_pState->fSunIntensity;
		PSCB.g_sunDirection = { cosf(m_pState->fSunAngle * 0.0174533f), 0, sinf(m_pState->fSunAngle * 0.0174533f) };
		PSCB.g_useMicrofacetFresnel = m_pState->bUseMicrofacetFresnel ? 1.0f : 0.0f;
		PSCB.g_useMicrofacetSpecular = m_pState->bUseMicrofacetSpecular ? 1.0f : 0.0f;
		PSCB.g_useMicrofacetReflection = m_pState->bUseMicrofacetReflection ? 1.0f : 0.0f;
		PSCB.g_showCascades = m_pState->bShowCascades ? 1.0f : 0.0f;
		PSCB.g_eyePos = { eyePoint.x, eyePoint.z, eyePoint.y };

		m_hCommandList->writeBuffer(m_hOceanSurfacePSCB, &PSCB, sizeof(PSCB));

		// Getting data for rendering the nodes from quadtree
		uint32_t numNodes;
		GFSDK_WaveWorks_Quadtree_NodeRenderingProperties* pNodes;
		std::vector<GFSDK_WaveWorks_Quadtree_NodeRenderingProperties> nodes;
		float cullMargin = GFSDK_WaveWorks_Wind_Waves_Simulation_GetConservativeMaxDisplacementEstimate(m_pState->hOceanWindSimulation);

		GFSDK_WaveWorks_Quadtree_GetNodesForRendering(m_pState->hOceanQuadtree, vm, projMatrix, viewportSize, cullMargin, &numNodes, &pNodes);
		nodes.assign(pNodes, pNodes + numNodes);

		// Getting the quadtree stats
		GFSDK_WaveWorks_Quadtree_GetStats(m_pState->hOceanQuadtree, m_pState->OceanQuadtreeStats);

		// We could render patches one by one, but it's much more draw calls and less efficient compared to instancing.
		// We can use instancing here, since there is just 16 possible patch topology types.
		// Up to 16 types of patch types (instance types) can be exposed by quadtree generator.
		
		// Start indices and index counts for patch types generated by quadtree
		std::vector<uint32_t> startIndices(16);
		std::vector<uint32_t> indexCounts(16);

		// Per-instance data for patch types
		std::vector <std::vector< OCEAN_VS_CBUFFER_PERINSTANCE_ENTRY > > perInstanceBuffers(16);

		// Reserving space for the entire constant buffer size (64kb or 4096 entries) in perInstanceBuffers
		// as current NVRHI implementation updates the entire contents of the constant buffers.
		for (uint32_t i = 0; i < 16; i++)
		{
			perInstanceBuffers[i].reserve(4096);
		}

		// Assembling the arrays
		for (uint32_t i = 0; i < nodes.size(); i++)
		{
			uint32_t patchType = nodes[i].patch_type;

			startIndices[patchType] = nodes[i].start_index;
			indexCounts[patchType] = nodes[i].num_indices;

			OCEAN_VS_CBUFFER_PERINSTANCE_ENTRY cbEntry;
			cbEntry.g_patchWorldspaceOrigin = nodes[i].patch_worldspace_origin;
			cbEntry.g_patchWorldspaceScale = nodes[i].patch_worldspace_scale;
			cbEntry.g_patchMorphConstantAndSign = nodes[i].geomorphing_distance_constant * nodes[i].geomorphing_sign;
			perInstanceBuffers[patchType].push_back(cbEntry);
		}

		// Rendering the quadtree patches using instancing
		for (uint32_t i = 0; i < 16; i++)
		{

			if (perInstanceBuffers[i].size() > 0)
			{
				// Writing the entire constant buffer
				m_hCommandList->writeBuffer(m_hOceanSurfacePerInstanceCB, (void*)perInstanceBuffers[i].data(), 4096 * sizeof(OCEAN_VS_CBUFFER_PERINSTANCE_ENTRY));

				// Setting up draw arguments to render up to 4096 instances.
				// Ocean quadtree usually generates <100 instances of patches of each type, 
				// but if there is a need to render more than 4096 instances, then structured / vertex buffer can be used instead of constant buffer,
				// or multiple draw calls with the constant buffer updates between them.
				nvrhi::DrawArguments da;
				da.vertexCount = indexCounts[i];
				da.startIndexLocation = startIndices[i];
				da.startVertexLocation = 0;
				da.startInstanceLocation = 0;
				da.instanceCount = std::min(4096, (int32_t)(perInstanceBuffers[i].size()));

				m_drawState.indexBuffer.handle = m_hOceanSurfaceIB;
				m_drawState.indexBuffer.format = nvrhi::Format::R32_UINT;
				m_drawState.indexBuffer.offset = 0;
				m_drawState.vertexBuffers.resize(1);
				m_drawState.vertexBuffers[0].buffer = m_hOceanSurfaceVB;
				m_drawState.vertexBuffers[0].slot = 0;
				m_drawState.vertexBuffers[0].offset = 0;

				m_hCommandList->setGraphicsState(m_drawState);
				m_hCommandList->drawIndexed(da);
			}
		}

		if (m_pState->bRenderMarkers)
		{
			RenderMarkers(FB);
		}
	}
	m_hCommandList->endTrackingTextureState(renderTarget, nvrhi::AllSubresources, nvrhi::ResourceStates::RENDER_TARGET);
	m_hCommandList->endTrackingTextureState(m_pState->hDepthBufferTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::DEPTH_WRITE);

	m_hCommandList->endTrackingTextureState(m_hWindWavesDisplacementsTextureArray, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
	m_hCommandList->endTrackingTextureState(m_hWindWavesGradientsTextureArray, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
	m_hCommandList->endTrackingTextureState(m_hWindWavesMomentsTextureArray, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
	m_hCommandList->endTrackingTextureState(m_hLocalWavesDisplacementsTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
	m_hCommandList->endTrackingTextureState(m_hLocalWavesGradientsTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
	m_hCommandList->close();
	m_pDevice->executeCommandList(m_hCommandList);
}

nvrhi::FramebufferHandle OceanRenderer::getFramebuffer(nvrhi::TextureHandle renderTarget, uint32_t swapChainIndex)
{
	nvrhi::TextureDesc td = renderTarget->GetDesc();

	// Reallocate the depth buffer if needed
	if (m_pState->hDepthBufferTexture == nullptr)
	{
		nvrhi::TextureDesc desc;
		desc.width = td.width;
		desc.height = td.height;
		desc.sampleCount = td.sampleCount;
		desc.debugName = "Depth target created in OceanRenderer::getFramebuffer";
		desc.format = nvrhi::Format::D24S8;
		desc.isRenderTarget = true;
		desc.clearValue = nvrhi::Color(1.0f);
		desc.useClearValue = true;
		desc.isTypeless = true; // Depth targets must be declared typeless in NVRHI
		m_pState->hDepthBufferTexture = m_pDevice->createTexture(desc);
		assert(m_pState->hDepthBufferTexture);
		m_hCommandList->beginTrackingTextureState(m_pState->hDepthBufferTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::COMMON);
		m_hCommandList->endTrackingTextureState(m_pState->hDepthBufferTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::DEPTH_WRITE);
	}

	// Create framebuffers if needed
	if (swapChainIndex >= m_hFBs.size()) m_hFBs.resize(swapChainIndex + 1);

	if (m_hFBs[swapChainIndex] == nullptr)
	{
		nvrhi::FramebufferDesc desc = nvrhi::FramebufferDesc()
			.addColorAttachment(renderTarget)
			.setDepthAttachment(m_pState->hDepthBufferTexture);

		m_hFBs[swapChainIndex] = m_pDevice->createFramebuffer(desc);
	}

	if (m_needToRecreatePSOs)
	{
		// Recreating PSOs
		nvrhi::GraphicsPipelineDesc markerPSOdesc = m_hMarkerPSO->GetDesc();
		m_hMarkerPSO = nullptr;
		m_hMarkerPSO = m_pDevice->createGraphicsPipeline(markerPSOdesc, m_hFBs[swapChainIndex]);

		nvrhi::GraphicsPipelineDesc solidPSOdesc = m_hOceanSurfaceSolidPSO->GetDesc();
		m_hOceanSurfaceSolidPSO = nullptr;
		m_hOceanSurfaceSolidPSO = m_pDevice->createGraphicsPipeline(solidPSOdesc, m_hFBs[swapChainIndex]);

		nvrhi::GraphicsPipelineDesc wireframePSOdesc = m_hOceanSurfaceWireframePSO->GetDesc();
		m_hOceanSurfaceWireframePSO = nullptr;
		m_hOceanSurfaceWireframePSO = m_pDevice->createGraphicsPipeline(wireframePSOdesc, m_hFBs[swapChainIndex]);
		m_needToRecreatePSOs = false;
	}

	assert(m_hFBs.at(swapChainIndex));
	return m_hFBs[swapChainIndex];
}

void OceanRenderer::RebindOceanResources()
{
	m_pDevice->waitForIdle();

	// Re-creating quadtree vertex and index buffers if needed
	if (m_pState->bNeedToUpdateQuadtreeProperties)
	{
		m_hOceanSurfaceVB = nullptr;
		m_hOceanSurfaceIB = nullptr;

		// Creating ocean surface vertex and index buffers
		uint32_t numVertices = 0;
		uint32_t numIndices = 0;
		gfsdk_float2 *pVertexData = nullptr;
		uint32_t *pIndexData = nullptr;
		std::vector<gfsdk_float2> oceanVertices;
		std::vector<uint32_t> oceanIndices;

		GFSDK_WaveWorks_Quadtree_GetGeometryData(m_pState->hOceanQuadtree, &numVertices, &pVertexData, &numIndices, &pIndexData);

		oceanVertices.assign(pVertexData, pVertexData + numVertices);
		oceanIndices.assign(pIndexData, pIndexData + numIndices);

		nvrhi::BufferDesc bd;
		bd.byteSize = numVertices * sizeof(gfsdk_float2);
		bd.isVertexBuffer = true;
		bd.isIndexBuffer = false;
		m_hOceanSurfaceVB = m_pDevice->createBuffer(bd);
		m_hCommandList->beginTrackingBufferState(m_hOceanSurfaceVB, nvrhi::ResourceStates::COMMON);
		m_hCommandList->writeBuffer(m_hOceanSurfaceVB, oceanVertices.data(), bd.byteSize);
		m_hCommandList->endTrackingBufferState(m_hOceanSurfaceVB, nvrhi::ResourceStates::VERTEX_BUFFER, true);

		bd.byteSize = numIndices * sizeof(uint32_t);
		bd.isVertexBuffer = false;
		bd.isIndexBuffer = true;
		m_hOceanSurfaceIB = m_pDevice->createBuffer(bd);
		m_hCommandList->beginTrackingBufferState(m_hOceanSurfaceIB, nvrhi::ResourceStates::COMMON);
		m_hCommandList->writeBuffer(m_hOceanSurfaceIB, oceanIndices.data(), bd.byteSize);
		m_hCommandList->endTrackingBufferState(m_hOceanSurfaceIB, nvrhi::ResourceStates::INDEX_BUFFER, true);
	}

	// Re-creating NVRHI textures from native objects exposed from WaveWorks
	m_hWindWavesDisplacementsTextureArray = nullptr;
	m_hWindWavesGradientsTextureArray = nullptr;
	m_hWindWavesMomentsTextureArray = nullptr;

	m_hLocalWavesDisplacementsTexture = nullptr;
	m_hLocalWavesGradientsTexture = nullptr;

	GFSDK_WaveWorks_Wind_Waves_Rendering_Data windWavesRenderingData;
	GFSDK_WaveWorks_Local_Waves_Rendering_Data localWavesRenderingData;
	GFSDK_WaveWorks_Wind_Waves_GetDataForRendering(m_pState->hOceanWindSimulation, windWavesRenderingData);
	GFSDK_WaveWorks_Local_Waves_GetDataForRendering(m_pState->hOceanLocalSimulation, localWavesRenderingData);

	nvrhi::TextureDesc texArrayDesc;
	texArrayDesc.width = windWavesRenderingData.size_of_texture_arrays;
	texArrayDesc.height = windWavesRenderingData.size_of_texture_arrays;
	texArrayDesc.format = nvrhi::Format::RGBA16_FLOAT;
	texArrayDesc.isUAV = true;
	texArrayDesc.dimension = nvrhi::TextureDimension::Texture2DArray;
	texArrayDesc.arraySize = 4;
	texArrayDesc.mipLevels = getNumMipLevels(windWavesRenderingData.size_of_texture_arrays);
	texArrayDesc.initialState = nvrhi::ResourceStates::SHADER_RESOURCE;
	texArrayDesc.keepInitialState = true;
	m_hWindWavesDisplacementsTextureArray = m_pDevice->createHandleForNativeTexture(NATIVE_OBJECT, windWavesRenderingData.displacements_texture_array, texArrayDesc);
	m_hWindWavesGradientsTextureArray = m_pDevice->createHandleForNativeTexture(NATIVE_OBJECT, windWavesRenderingData.gradients_texture_array, texArrayDesc);
	m_hWindWavesMomentsTextureArray = m_pDevice->createHandleForNativeTexture(NATIVE_OBJECT, windWavesRenderingData.moments_texture_array, texArrayDesc);

	nvrhi::TextureDesc texDesc;
	texDesc.width = localWavesRenderingData.size_of_texture;
	texDesc.height = localWavesRenderingData.size_of_texture;
	texDesc.format = nvrhi::Format::RGBA16_FLOAT;
	texDesc.isUAV = true;
	texDesc.dimension = nvrhi::TextureDimension::Texture2D;
	texDesc.mipLevels = getNumMipLevels(localWavesRenderingData.size_of_texture);
	texDesc.initialState = nvrhi::ResourceStates::SHADER_RESOURCE;
	texDesc.keepInitialState = true;
	m_hLocalWavesDisplacementsTexture = m_pDevice->createHandleForNativeTexture(NATIVE_OBJECT, localWavesRenderingData.displacements_texture, texDesc);
	m_hLocalWavesGradientsTexture = m_pDevice->createHandleForNativeTexture(NATIVE_OBJECT, localWavesRenderingData.gradients_texture, texDesc);
	
	// Updating binding set
	nvrhi::TextureDesc tdWindDisplacements = m_hWindWavesDisplacementsTextureArray->GetDesc();
	nvrhi::TextureDesc tdLocalDisplacements = m_hLocalWavesDisplacementsTexture->GetDesc();

	nvrhi::BindingSetDesc bsDesc;
	bsDesc.VS = { nvrhi::BindingSetItem::ConstantBuffer(0, m_hOceanSurfacePerInstanceCB ),
					nvrhi::BindingSetItem::ConstantBuffer(1, m_hOceanSurfaceVSHSDSCB ) };
	bsDesc.HS = { nvrhi::BindingSetItem::ConstantBuffer(0, m_hOceanSurfaceVSHSDSCB ) };
	bsDesc.DS = { nvrhi::BindingSetItem::Texture_SRV(0, m_hWindWavesDisplacementsTextureArray, nvrhi::Format::RGBA16_FLOAT, nvrhi::TextureSubresourceSet(0,tdWindDisplacements.mipLevels,0,tdWindDisplacements.arraySize) ),
					nvrhi::BindingSetItem::Texture_SRV(1, m_hLocalWavesDisplacementsTexture, nvrhi::Format::RGBA16_FLOAT),
					nvrhi::BindingSetItem::Sampler(0, m_hBilinearSampler ),
					nvrhi::BindingSetItem::Sampler(1, m_hBilinearClampToEdgeSampler ),
					nvrhi::BindingSetItem::ConstantBuffer(0, m_hOceanSurfaceVSHSDSCB) };
	bsDesc.PS = { nvrhi::BindingSetItem::Texture_SRV(0, m_hWindWavesGradientsTextureArray, nvrhi::Format::RGBA16_FLOAT, nvrhi::TextureSubresourceSet(0,tdWindDisplacements.mipLevels,0,tdWindDisplacements.arraySize)),
					nvrhi::BindingSetItem::Texture_SRV(1, m_hWindWavesMomentsTextureArray, nvrhi::Format::RGBA16_FLOAT, nvrhi::TextureSubresourceSet(0,tdWindDisplacements.mipLevels,0,tdWindDisplacements.arraySize)),
					nvrhi::BindingSetItem::Texture_SRV(2, m_hLocalWavesDisplacementsTexture,  nvrhi::Format::RGBA16_FLOAT),
					nvrhi::BindingSetItem::Texture_SRV(3, m_hLocalWavesGradientsTexture,  nvrhi::Format::RGBA16_FLOAT),
					nvrhi::BindingSetItem::Texture_SRV(4, m_hFoamIntensityTexture, nvrhi::Format::RGBA8_UNORM),
					nvrhi::BindingSetItem::Texture_SRV(5, m_hFoamBubblesTexture,  nvrhi::Format::RGBA8_UNORM),
					nvrhi::BindingSetItem::Texture_SRV(6, m_hWindGustsTexture,  nvrhi::Format::RGBA8_UNORM),
					nvrhi::BindingSetItem::Texture_SRV(7, m_pState->hDynamicSkyDomeTexture,  nvrhi::Format::RGBA16_FLOAT),
					nvrhi::BindingSetItem::Sampler(0, m_hBilinearSampler),
					nvrhi::BindingSetItem::Sampler(1, m_hAnisotropicSampler),
					nvrhi::BindingSetItem::Sampler(2, m_hBilinearClampToEdgeSampler ),
					nvrhi::BindingSetItem::ConstantBuffer(0, m_hOceanSurfacePSCB ) };
	m_hOceanSurfaceBindings = m_pDevice->createBindingSet(bsDesc, m_hOceanSurfaceBindingLayout);

}

void OceanRenderer::Shutdown()
{
	free(m_pReadbackPositions);
	free(m_pReadbackResults);
	m_pState->hDepthBufferTexture = nullptr;
}

bool OceanRenderer::KeyboardUpdate(int key, int scancode, int action, int mods)
{
	m_pState->Camera.KeyboardUpdate(key, scancode, action, mods);
	return false;
}

bool OceanRenderer::MousePosUpdate(double xpos, double ypos)
{
	m_pState->Camera.MousePosUpdate(xpos, ypos);
	return false;
}

bool OceanRenderer::MouseButtonUpdate(int button, int action, int mods)
{
	m_pState->Camera.MouseButtonUpdate(button, action, mods);
	return false;
}

void OceanRenderer::Animate(float elapsedTimeSeconds)
{
	m_pState->Camera.Animate(elapsedTimeSeconds);

	// Adding rain
	if ((m_pState->bRain) && (!m_pState->bNeedToUpdateLocalWavesSimulationProperties))
	{
		gfsdk_float4 dd[200];
		for (int i = 0; i < 200; i++)
		{
			uint32_t x = rand() % m_pState->OceanLocalSimulationSettings.simulation_domain_grid_size;
			uint32_t y = rand() % m_pState->OceanLocalSimulationSettings.simulation_domain_grid_size;
			dd[i].x = (float)x;
			dd[i].y = (float)y;
			dd[i].z = m_pState->fRainDropSize * (rand()%100)*0.01f;
			dd[i].w = 0.0f;
		}
		GFSDK_WaveWorks_Local_Waves_Simulation_AddDisturbances(m_pState->hOceanLocalSimulation, dd, 200);
	}

	// Adding "boat"
	if ((m_pState->bBoat) && (!m_pState->bNeedToUpdateLocalWavesSimulationProperties))
	{
		float time = (float)m_pState->dSimulationTime;
		uint32_t N = m_pState->OceanLocalSimulationSettings.simulation_domain_grid_size;
		m_pState->fBoatX = (float)(100.0f * sinf(time * 0.1f));
		m_pState->fBoatY = (float)(100.0f*cosf(time * 0.13f));
		float dX = 0.1f * 100.0f*cosf((float)time * 0.1f);
		float dY = -0.13f * 100.0f*sinf((float)time * 0.13f);
		float speed = sqrtf(dX*dX + dY*dY);
		m_pState->fBoatX = (float)(100.0f*sinf(time * 0.1f));
		m_pState->fBoatY = (float)(100.0f*cosf(time * 0.13f));
		gfsdk_float4 dd[100];
		for (int32_t i = 0; i < 10; i++)
		{
			for (int32_t j = 0; j < 10; j++)
			{
				float x = m_pState->fBoatX + i - (10 / 2) + (N / 2);
				float y = m_pState->fBoatY + j - (10 / 2) + (N / 2);
				dd[i + 10 * j].x = x;
				dd[i + 10 * j].y = y;
				dd[i + 10 * j].z = 0.13f*speed*elapsedTimeSeconds;
				dd[i + 10 * j].w = 0;
			}
		}
		GFSDK_WaveWorks_Local_Waves_Simulation_AddDisturbances(m_pState->hOceanLocalSimulation, dd, 100);
		gfsdk_float2 cameraXY = { m_pState->fBoatX, m_pState->fBoatY };
		cameraXY *= m_pState->OceanLocalSimulationSettings.simulation_domain_worldspace_size / m_pState->OceanLocalSimulationSettings.simulation_domain_grid_size;
		gfsdk_float3 cameraPos = { cameraXY.x - 30.0f , 10, cameraXY.y - 3.0f};
		gfsdk_float3 atPos = { cameraXY.x , 5, cameraXY.y };
		m_pState->Camera.LookAt(cameraPos, atPos);
	}

	// Simulating local waves
	if (m_pState->bNeedToResetLocalWavesSimulation)
	{
		GFSDK_WaveWorks_Local_Waves_Simulation_Reset(m_pState->hOceanLocalSimulation);
		m_pState->bNeedToResetLocalWavesSimulation = false;
	}

	if (m_pState->bSimulateWater || m_pState->bForceKick || (gfsdk_wwresult_NONE == GFSDK_WaveWorks_Local_Waves_Simulation_GetStagingCursor(m_pState->hOceanLocalSimulation, NULL)))
	{
		GFSDK_WaveWorks_Local_Waves_Simulation_SetDeltaTime(m_pState->hOceanLocalSimulation, elapsedTimeSeconds);
		GFSDK_WaveWorks_Local_Waves_Simulation_Kick(m_pState->hOceanLocalSimulation, &m_pState->iLastLocalSimulationKickID);
	}

	if (m_pState->SyncMode >= SynchronizationMode_RenderOnly)
	{
		// Block until the just-submitted kick is ready to render
		uint64_t stagingCursorKickID = m_pState->iLastLocalSimulationKickID - 1;	// Just ensure that the initial value is different from last kick,
																				// so that we continue waiting if the staging cursor is empty
		GFSDK_WaveWorks_Local_Waves_Simulation_GetStagingCursor(m_pState->hOceanLocalSimulation, &stagingCursorKickID);
		while (stagingCursorKickID != m_pState->iLastLocalSimulationKickID)
		{
			const bool doBlock = true;
			GFSDK_WaveWorks_Local_Waves_Simulation_AdvanceStagingCursor(m_pState->hOceanLocalSimulation, doBlock);
			GFSDK_WaveWorks_Local_Waves_Simulation_GetStagingCursor(m_pState->hOceanLocalSimulation, &stagingCursorKickID);
		}

		if ((m_pState->SyncMode >= SynchronizationMode_Readback) && (m_pState->iReadbackUsage > 0))
		{
			uint64_t readbackCursorKickID = m_pState->iLastLocalSimulationKickID - 1;	// Just ensure that the initial value is different from last kick,
																						// so that we continue waiting if the staging cursor is empty
			while (readbackCursorKickID != m_pState->iLastLocalSimulationKickID)
			{
				const bool doBlock = true;
				GFSDK_WaveWorks_Local_Waves_Simulation_AdvanceReadbackCursor(m_pState->hOceanLocalSimulation, doBlock);
				GFSDK_WaveWorks_Local_Waves_Simulation_GetReadbackCursor(m_pState->hOceanLocalSimulation, &readbackCursorKickID);
			}
		}
	}
	else
	{
		
		// Keep feeding the simulation pipeline until it is full - this loop should skip in all
		// cases except the first iteration, when the simulation pipeline is first 'primed'
		while (GFSDK_WaveWorks_Local_Waves_Simulation_GetStagingCursor(m_pState->hOceanLocalSimulation, NULL) == gfsdk_wwresult_NONE)
		{
			GFSDK_WaveWorks_Local_Waves_Simulation_SetDeltaTime(m_pState->hOceanLocalSimulation, elapsedTimeSeconds);
			GFSDK_WaveWorks_Local_Waves_Simulation_Kick(m_pState->hOceanLocalSimulation, &m_pState->iLastLocalSimulationKickID);
		}
		
	}
	
	// Getting simulation timings
	GFSDK_WaveWorks_Local_Waves_Simulation_GetStats(m_pState->hOceanLocalSimulation, m_pState->OceanLocalSimulationStats);

	// Exponential filtering for stats
	m_pState->OceanLocalSimulationStatsFiltered.CPU_main_thread_wait_time = m_pState->OceanLocalSimulationStatsFiltered.CPU_main_thread_wait_time*0.999f + 0.001f*m_pState->OceanLocalSimulationStats.CPU_main_thread_wait_time;
	m_pState->OceanLocalSimulationStatsFiltered.CPU_threads_total_time = m_pState->OceanLocalSimulationStatsFiltered.CPU_threads_total_time*0.999f + 0.001f*m_pState->OceanLocalSimulationStats.CPU_threads_total_time;
	m_pState->OceanLocalSimulationStatsFiltered.GPU_simulation_time = m_pState->OceanLocalSimulationStatsFiltered.GPU_simulation_time*0.999f + 0.001f*m_pState->OceanLocalSimulationStats.GPU_simulation_time;
	m_pState->OceanLocalSimulationStatsFiltered.GPU_update_time = m_pState->OceanLocalSimulationStatsFiltered.GPU_update_time*0.999f + 0.001f*m_pState->OceanLocalSimulationStats.GPU_update_time;
	m_pState->OceanLocalSimulationStatsFiltered.GPU_total_time = m_pState->OceanLocalSimulationStatsFiltered.GPU_total_time*0.999f + 0.001f*m_pState->OceanLocalSimulationStats.GPU_total_time;

	// deduce the rendering latency of the WaveWorks local waves simulation pipeline
	{
		uint64_t staging_cursor_kickID = 0;
		GFSDK_WaveWorks_Local_Waves_Simulation_GetStagingCursor(m_pState->hOceanLocalSimulation, &staging_cursor_kickID);
		m_pState->iLocalSimulationRenderLatency = (uint32_t)(m_pState->iLastLocalSimulationKickID - staging_cursor_kickID);
	}

	// likewise with the readback latency
	if (m_pState->iReadbackUsage > 0)
	{
		uint64_t readback_cursor_kickID = 0;
		if (GFSDK_WaveWorks_Local_Waves_Simulation_GetReadbackCursor(m_pState->hOceanLocalSimulation, &readback_cursor_kickID) == gfsdk_wwresult_OK)
		{
			m_pState->iLocalSimulationReadbackLatency = (int32_t)(m_pState->iLastLocalSimulationKickID - readback_cursor_kickID);
		}
		else
		{
			m_pState->iLocalSimulationReadbackLatency = -1;
		}
	}
	else
	{
		m_pState->iLocalSimulationReadbackLatency = -1;
	}



	// Simulating wind waves
	if (m_pState->bSimulateWater || m_pState->bForceKick || (gfsdk_wwresult_NONE == GFSDK_WaveWorks_Wind_Waves_Simulation_GetStagingCursor(m_pState->hOceanWindSimulation, NULL)))
	{
		m_pState->dSimulationTime += elapsedTimeSeconds;
		GFSDK_WaveWorks_Wind_Waves_Simulation_SetTime(m_pState->hOceanWindSimulation, m_pState->dSimulationTime);
		GFSDK_WaveWorks_Wind_Waves_Simulation_Kick(m_pState->hOceanWindSimulation, &m_pState->iLastWindSimulationKickID);
	}


	if (m_pState->SyncMode >= SynchronizationMode_RenderOnly)
	{
		// Block until the just-submitted kick is ready to render
		uint64_t stagingCursorKickID = m_pState->iLastWindSimulationKickID - 1;	// Just ensure that the initial value is different from last kick,
															// so that we continue waiting if the staging cursor is empty
		GFSDK_WaveWorks_Wind_Waves_Simulation_GetStagingCursor(m_pState->hOceanWindSimulation, &stagingCursorKickID);
		while (stagingCursorKickID != m_pState->iLastWindSimulationKickID)
		{
			const bool doBlock = true;
			GFSDK_WaveWorks_Wind_Waves_Simulation_AdvanceStagingCursor(m_pState->hOceanWindSimulation, doBlock);
			GFSDK_WaveWorks_Wind_Waves_Simulation_GetStagingCursor(m_pState->hOceanWindSimulation, &stagingCursorKickID);
		}

		if ((m_pState->SyncMode >= SynchronizationMode_Readback) && (m_pState->iReadbackUsage > 0))
		{
			uint64_t readbackCursorKickID = m_pState->iLastWindSimulationKickID - 1;	// Just ensure that the initial value is different from last kick,
																// so that we continue waiting if the staging cursor is empty
			while (readbackCursorKickID != m_pState->iLastWindSimulationKickID)
			{
				const bool doBlock = true;
				GFSDK_WaveWorks_Wind_Waves_Simulation_AdvanceReadbackCursor(m_pState->hOceanWindSimulation, doBlock);
				GFSDK_WaveWorks_Wind_Waves_Simulation_GetReadbackCursor(m_pState->hOceanWindSimulation, &readbackCursorKickID);
			}
		}
	}
	else
	{
		// Keep feeding the simulation pipeline until it is full - this loop should skip in all
		// cases except the first iteration, when the simulation pipeline is first 'primed'
		while (GFSDK_WaveWorks_Wind_Waves_Simulation_GetStagingCursor(m_pState->hOceanWindSimulation, NULL) == gfsdk_wwresult_NONE)
		{
			GFSDK_WaveWorks_Wind_Waves_Simulation_SetTime(m_pState->hOceanWindSimulation, m_pState->dSimulationTime);
			GFSDK_WaveWorks_Wind_Waves_Simulation_Kick(m_pState->hOceanWindSimulation, &m_pState->iLastWindSimulationKickID);
		}
	}

	m_pState->bForceKick = false;

	// Exercise the readback archiving API
	if (GFSDK_WaveWorks_Wind_Waves_Simulation_GetReadbackCursor(m_pState->hOceanWindSimulation, &m_pState->iLastWindSimulationReadbackKickID) == gfsdk_wwresult_OK)
	{
		if ((m_pState->iLastWindSimulationReadbackKickID - m_pState->iLastWindSimulationArchivedKickID) > m_pState->ReadbackArchiveInterval)
		{
			GFSDK_WaveWorks_Wind_Waves_Simulation_ArchiveDisplacements(m_pState->hOceanWindSimulation);
			m_pState->iLastWindSimulationArchivedKickID = m_pState->iLastWindSimulationReadbackKickID;
		}
	}

	// deduce the rendering latency of the WaveWorks wind waves simulation pipeline
	{
		uint64_t staging_cursor_kickID = 0;
		GFSDK_WaveWorks_Wind_Waves_Simulation_GetStagingCursor(m_pState->hOceanWindSimulation, &staging_cursor_kickID);
		m_pState->iWindSimulationRenderLatency = (uint32_t)(m_pState->iLastWindSimulationKickID - staging_cursor_kickID);
	}

	// likewise with the readback latency
	if (m_pState->iReadbackUsage > 0)
	{
		uint64_t readback_cursor_kickID = 0;
		if (GFSDK_WaveWorks_Wind_Waves_Simulation_GetReadbackCursor(m_pState->hOceanWindSimulation, &readback_cursor_kickID) == gfsdk_wwresult_OK)
		{
			m_pState->iWindSimulationReadbackLatency = (int32_t)(m_pState->iLastWindSimulationKickID - readback_cursor_kickID);
		}
		else
		{
			m_pState->iWindSimulationReadbackLatency = -1;
		}
	}
	else
	{
		m_pState->iWindSimulationReadbackLatency = -1;
	}

	// getting simulation timings
	GFSDK_WaveWorks_Wind_Waves_Simulation_GetStats(m_pState->hOceanWindSimulation, m_pState->OceanWindSimulationStats);

	// exponential filtering for stats
	m_pState->OceanWindSimulationStatsFiltered.CPU_main_thread_wait_time = m_pState->OceanWindSimulationStatsFiltered.CPU_main_thread_wait_time*0.999f + 0.001f*m_pState->OceanWindSimulationStats.CPU_main_thread_wait_time;
	m_pState->OceanWindSimulationStatsFiltered.CPU_threads_total_time = m_pState->OceanWindSimulationStatsFiltered.CPU_threads_total_time*0.999f + 0.001f*m_pState->OceanWindSimulationStats.CPU_threads_total_time;
	m_pState->OceanWindSimulationStatsFiltered.GPU_simulation_time = m_pState->OceanWindSimulationStatsFiltered.GPU_simulation_time*0.999f + 0.001f*m_pState->OceanWindSimulationStats.GPU_simulation_time;
	m_pState->OceanWindSimulationStatsFiltered.GPU_update_time = m_pState->OceanWindSimulationStatsFiltered.GPU_update_time*0.999f + 0.001f*m_pState->OceanWindSimulationStats.GPU_update_time;
	m_pState->OceanWindSimulationStatsFiltered.GPU_total_time = m_pState->OceanWindSimulationStatsFiltered.GPU_total_time*0.999f + 0.001f*m_pState->OceanWindSimulationStats.GPU_total_time;

	// reading back marker coords
	if (m_pState->iReadbackUsage > 0)
	{
		UpdateMarkers();
	}
	
}

void OceanRenderer::ResetCameraView()
{
	gfsdk_float3 eyePt = { -50.0f, 20.0f, 30.0f };
	gfsdk_float3 lookAtPt = { 0.0f, 20.0f, 0.0f };
	m_pState->Camera.LookAt(eyePt, lookAtPt);

	m_pState->Camera.SetSpeeds(5.0f, 0.005f);
	updateCameraProjection();
}

void OceanRenderer::UpdateMarkers()
{
	
	float distance_between_markers = 2.0f;// *m_pState->OceanWindSimulationSettings.simulation_period / (float)NumMarkersXY;
	float center_x = 0;
	float center_y = 0;
	
	for (int i = 0; i < NumMarkersXY; i++)
		for (int j = 0; j < NumMarkersXY; j++)

	{
			float x = center_x - distance_between_markers*(i - NumMarkersXY / 2);
			float y = center_y - distance_between_markers*(j - NumMarkersXY / 2);
			m_pReadbackPositions[i + j*NumMarkersXY].x = x;
			m_pReadbackPositions[i + j*NumMarkersXY].y = y;
	}
	

	m_pState->fMinDisplacement = 0;
	m_pState->fMaxDisplacement = 0;
	
	/*
	for (int p = 0; p < m_pState->iRefinementSteps; p++)
	{
		GFSDK_WaveWorks_Simulation_GetDisplacements(m_pState->hOceanSimulationForPhysics, m_pReadbackPositions, m_pReadbackResults, NumMarkers, m_pState->iReadbackUsage == 2 ? true : false);
		for (int i = 0; i < NumMarkersXY; i++)
			for (int j = 0; j < NumMarkersXY; j++)

			{
				float x = center_x - distance_between_markers*(i - NumMarkersXY / 2);
				float y = center_y - distance_between_markers*(j - NumMarkersXY / 2);
				m_pReadbackPositions[i + j*NumMarkersXY].x = x;
				m_pReadbackPositions[i + j*NumMarkersXY].y = y;
				m_pReadbackPositions[i + j*NumMarkersXY].x -= m_pReadbackResults[i + j*NumMarkersXY].x;
				m_pReadbackPositions[i + j*NumMarkersXY].y -= m_pReadbackResults[i + j*NumMarkersXY].y;

			}
	}

	for (int i = 0; i < NumMarkersXY; i++)
		for (int j = 0; j < NumMarkersXY; j++)
		{
			float x = center_x - distance_between_markers*(i - NumMarkersXY / 2);
			float y = center_y - distance_between_markers*(j - NumMarkersXY / 2);
			m_pReadbackResults[i + j*NumMarkersXY].x = x;
			m_pReadbackResults[i + j*NumMarkersXY].y = y;
		}
		*/
	int k = 1;
	for (int i = 0; i < k; i++)
	{

		//GFSDK_WaveWorks_Local_Waves_Simulation_GetDisplacements(m_pState->hOceanLocalSimulation, &m_pReadbackPositions[(NumMarkers*i) / k], &m_pReadbackResults[(NumMarkers*i) / k], NumMarkers / k, m_pState->iReadbackUsage == 2 ? true : false);

		GFSDK_WaveWorks_Wind_Waves_Simulation_GetDisplacements(m_pState->hOceanWindSimulation, &m_pReadbackPositions[(NumMarkers*i)/k], &m_pReadbackResults[(NumMarkers*i) / k], NumMarkers/k, m_pState->iReadbackUsage == 2 ? true : false);
		//GFSDK_WaveWorks_Wind_Waves_Simulation_GetArchivedDisplacements(m_pState->hOceanWindSimulation, m_pState->fReadbackCoord, &m_pReadbackPositions[(NumMarkers*i) / k], &m_pReadbackResults[(NumMarkers*i) / k], NumMarkers / k);
	}
	
	float variance_sqr = 0;
	for (int i = 0; i < NumMarkers; i++)
	{
		m_pReadbackResults[i].x += m_pReadbackPositions[i].x;
		m_pReadbackResults[i].y += m_pReadbackPositions[i].y;
		if (m_pReadbackResults[i].z < m_pState->fMinDisplacement) m_pState->fMinDisplacement = m_pReadbackResults[i].z;
		if (m_pReadbackResults[i].z > m_pState->fMaxDisplacement) m_pState->fMaxDisplacement = m_pReadbackResults[i].z;
		variance_sqr += m_pReadbackResults[i].z*m_pReadbackResults[i].z;
	}
	if (m_pState->fMaxTotalDisplacement < m_pState->fMaxDisplacement) m_pState->fMaxTotalDisplacement = m_pState->fMaxDisplacement;
	if (m_pState->fMinTotalDisplacement > m_pState->fMinDisplacement) m_pState->fMinTotalDisplacement = m_pState->fMinDisplacement;
	variance_sqr /= NumMarkers;
	m_pState->fSWH = 4.0f * sqrtf(variance_sqr);
	
	
}

void OceanRenderer::RenderMarkers(nvrhi::FramebufferHandle hFramebuffer)
{
	// Set initial resource states (command list has been opened already in OceanRenderer::Render)
	m_hCommandList->beginTrackingBufferState(m_hMarkerVSCB, nvrhi::ResourceStates::CONSTANT_BUFFER);
	m_hCommandList->beginTrackingBufferState(m_hMarkerInstanceBuffer, nvrhi::ResourceStates::VERTEX_BUFFER);

	// Initialize sky draw state
	nvrhi::Viewport vp;
	vp.maxX = float(m_backBufferWidth);
	vp.maxY = float(m_backBufferHeight);

	nvrhi::Rect scissor;
	scissor.maxX = m_backBufferWidth;
	scissor.maxY = m_backBufferHeight;

	m_drawState.viewport.viewports.resize(1);
	m_drawState.viewport.viewports[0] = vp;
	m_drawState.viewport.scissorRects = { scissor };

	// Update constant buffers
	MARKER_VS_CBUFFER VSCB;

	gfsdk_float4x4 viewProjMatrix = m_pState->Camera.GetWorldToClip();
	gfsdk_float3 eyePoint = m_pState->Camera.GetCameraPosition();

	VSCB.g_matViewProj = { viewProjMatrix._11, viewProjMatrix._12, viewProjMatrix._13, viewProjMatrix._14,
		viewProjMatrix._31, viewProjMatrix._32, viewProjMatrix._33,viewProjMatrix._34,
		viewProjMatrix._21, viewProjMatrix._22, viewProjMatrix._23,viewProjMatrix._24,
		viewProjMatrix._41, viewProjMatrix._42, viewProjMatrix._43,viewProjMatrix._44 };

	m_hCommandList->writeBuffer(m_hMarkerVSCB, &VSCB, sizeof(VSCB));

	// Uploading the readback results to GPU
	gfsdk_float4* uploadData = (gfsdk_float4*)m_pDevice->mapBuffer(m_hMarkerInstanceUploadBuffers[m_markerInstanceUploadBufferNumber], nvrhi::CpuAccessMode::Write);
	memcpy(uploadData, m_pReadbackResults, NumMarkers * sizeof(gfsdk_float4));
	m_pDevice->unmapBuffer(m_hMarkerInstanceUploadBuffers[m_markerInstanceUploadBufferNumber]);

	m_hCommandList->copyBuffer(m_hMarkerInstanceBuffer, 0, m_hMarkerInstanceUploadBuffers[m_markerInstanceUploadBufferNumber], 0, NumMarkers * sizeof(gfsdk_float4));

	m_markerInstanceUploadBufferNumber ++;
	m_markerInstanceUploadBufferNumber %= 3;

	// Rendering
	nvrhi::DrawArguments da;
	da.vertexCount = 12;
	da.instanceCount = NumMarkers;
	da.startIndexLocation = 0;
	da.startVertexLocation = 0;

	m_drawState.indexBuffer.handle = m_hMarkerIB;
	m_drawState.indexBuffer.format = nvrhi::Format::R32_UINT;
	m_drawState.indexBuffer.offset = 0;
	m_drawState.vertexBuffers.resize(2);
	m_drawState.vertexBuffers[0].buffer = m_hMarkerVB;
	m_drawState.vertexBuffers[0].slot = 0;
	m_drawState.vertexBuffers[0].offset = 0;
	m_drawState.vertexBuffers[1].buffer = m_hMarkerInstanceBuffer;
	m_drawState.vertexBuffers[1].slot = 1;
	m_drawState.vertexBuffers[1].offset = 0;

	m_drawState.framebuffer = hFramebuffer;
	m_drawState.pipeline = m_hMarkerPSO;
	m_drawState.bindings = { m_hMarkerBindings };
	m_hCommandList->setGraphicsState(m_drawState);
	m_hCommandList->drawIndexed(da);

	// Set required resource states (command list will be closed in OceanRenderer::Render)
	m_hCommandList->endTrackingBufferState(m_hMarkerVSCB, nvrhi::ResourceStates::CONSTANT_BUFFER);
	m_hCommandList->endTrackingBufferState(m_hMarkerInstanceBuffer, nvrhi::ResourceStates::VERTEX_BUFFER);
}
